/* 
   MD 2021-03-29 testovane s aktualne najnovsou verziou OpenSSL v 1.1.1k
   MD 2018-11-06 upravene pre linkovanie s OpenSSL 1.1.1
                 (pouzita metoda TLS_client_method)
   MD 2018-11-02 vlozene upravy na odstraneie varovania a autentizaciu servera 
                 vlozene upravy na realizaciu autentizacie klienta (nacitanie potrebnych certifikatov 
                 a kluca)
                 odstranenie varovania o poradi hlavickovych suborov
                 (vid. komentare // MD)
*/

/** SSL/TLS Client
 * SSL/TLS client demonstration. This source code is cross-plateforme Windows and Linux.
 * Compile under Linux with : g++ main.cpp -Wall -lssl -lcrypto -o main
 * Usage :
 *  # run the client to 127.0.0.1 on port 1337 for SSLv2&3 protocol
 *      $ [./]client[.exe] 127.0.0.1 1337
 *  # run the client to 127.0.0.1 on port 1337 for TLSv1 protocol
 *      $ [./]client[.exe] 127.0.0.1 1337 1
 * @author x@s
 */

#define DEFAULT_PORT 443

#ifdef __unix__         // __unix__ is usually defined by compilers targeting Unix systems
#   include <unistd.h>
#   include <sys/socket.h>
#   include <resolv.h>
#   include <netdb.h>
#   define SOCKLEN_T socklen_t
#   define CLOSESOCKET close
#elif defined _WIN32    // _Win32 is usually defined by compilers targeting 32 or 64 bit Windows systems
#define WIN32_LEAN_AND_MEAN    // MD odstrani varovanie o potrebe prehodit poradie hlavickovych suborov
#   include <windows.h>
#   include <winsock2.h>
#   define SOCKLEN_T int
#   define CLOSESOCKET closesocket
#endif

#include <stdio.h>
#include <errno.h>
#include <malloc.h>
#include <string.h>

#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#ifdef _WIN32
WSADATA wsa;    // Winsock data
#endif

/**
 * printUsage function who describe the utilisation of this script.
 * @param char* bin : the name of the current binary.
 */
void printHeader(char* bin){
    printf("[?] Usage : %s <hostname> <port> [<method>]\n", bin);
    printf("[?] With optional <method> :\n");
    printf("\t1 :\tTLS method (default)");
// MD    printf("\t2 :\tSSL v2 (deprecated so disabled)\n");
// MD    printf("\t3 :\tSSL v3\n");
// MD    printf("\t4 :\tSSL v2 & v3 (default)\n");
    return;
}

/**
 * makeClientSocket function who create a traditionnal client socket to the hostname throught the port.
 * @param char* hostname : the target to connect to
 * @param int port : the port to connect throught
 * @return int socket ; the socket number created
 */
int makeClientSocket(const char *hostname, int port){
    int                 sock;
    struct hostent      *host;
    struct sockaddr_in  addr;
    #ifdef _WIN32
    WSAStartup(MAKEWORD(2,0),&wsa);
    #endif
    if((host = gethostbyname(hostname)) == NULL ){
        perror(hostname);
        abort();
    }
    sock = socket(PF_INET, SOCK_STREAM, 0);
    memset(&addr, 0, sizeof(addr));
    addr.sin_family         = AF_INET;
    addr.sin_port           = htons(port);
    addr.sin_addr.s_addr    = *(long*)(host->h_addr);
    if(connect(sock, (struct sockaddr*)&addr, sizeof(addr)) != 0){
        CLOSESOCKET(sock);
        perror(hostname);
        abort();
    }
    return sock;
}

/**
 * initSSLContext function who initialize the SSL/TLS engine with right method/protocol
 * @param int ctxMethod : the number coresponding to the method/protocol to use
 * @return SSL_CTX *ctx ; a pointer to the SSL context created
 */
SSL_CTX* initSSLContext(int ctxMethod){
    const SSL_METHOD *method;
    SSL_CTX *ctx;

    SSL_library_init();                     // initialize the SSL library
    SSL_load_error_strings();			    // bring in and register error messages
    OpenSSL_add_all_algorithms();		    // load usable algorithms

    switch(ctxMethod){                      // create new client-method instance
        case 1 :
            method = TLS_client_method();
            printf("[+] Use TLS method.\n");
            break;
        default :
            method = TLS_client_method();
            printf("[+] Use TLS method.\n");
    }

    ctx = SSL_CTX_new(method);			    // create new context from selected method
    if(ctx == NULL){
        ERR_print_errors_fp(stderr);
        abort();
    }

// MD zabezpecienie overenia certifikatu servera pomocou CA
if (SSL_CTX_load_verify_locations(ctx, "myCA.pem", 0)) {
   printf("\nCA certificate loaded\n");
   SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER, NULL);
}
else {
   printf("\nCA certificate not loaded! Abort ...\n");
   abort();

}


// MD nacitanie dat pre autentizaciu klineta
#if (1)
int res = 0;


res = SSL_CTX_use_certificate_file(ctx, "client.pem", SSL_FILETYPE_PEM);
if (res <= 0) {
    // handle error
   printf("\nCLIENT certificate not loaded! Abort ...\n");
   abort();
}
res = SSL_CTX_use_PrivateKey_file(ctx, "client.key", SSL_FILETYPE_PEM);
if (res <= 0) {
    // handle error

   printf("\nCLIENT key not loaded! Abort ...\n");
   abort();
}

 /* verify private key */
    if ( !SSL_CTX_check_private_key(ctx) )
    {
        fprintf(stderr, "Private key does not match the public certificate\n");
        abort();
    }
    printf("Certificate attached.\n");
#endif


    return ctx;
}

/**
 * showCerts function who catch and print out certificat's data from the server
 * @param SSL* ssl : the SSL/TLS connection
 */
void showCerts(SSL* ssl){
    X509 *cert;
    char *subject, *issuer;

    cert = SSL_get_peer_certificate(ssl);   // get the server's certificate
    if(cert != NULL){
        subject = X509_NAME_oneline(X509_get_subject_name(cert), 0, 0); // get certificat's subject
        issuer  = X509_NAME_oneline(X509_get_issuer_name(cert), 0, 0);  // get certificat's issuer

        printf("[+] Server certificates :\n");
        printf("\tSubject: %s\n", subject);
        printf("\tIssuer: %s\n", issuer);

        free(subject);                      // free the malloc'ed string
        free(issuer);					    // free the malloc'ed string
        X509_free(cert);					// free the malloc'ed certificate copy

        if(SSL_get_verify_result(ssl) == X509_V_OK) // check certificat's trust
            printf("[+] Server certificates X509 is trust!\n");
        else
            printf("[-] Server certificates X509 is not trust...\n");
    }
    else
        printf("[-] No server's certificates\n");
    return;
}

/**
 * main function who coordinate the socket and SSL connection creation, then receive and emit data to and from the server.
 */
int main(int argc, char **argv){
    int     sock, bytes, ctxMethod, port;
    SSL_CTX *ctx;
    SSL     *ssl;
    char    buf[1024];
    char    *hostname;

    if(argc != 3 && argc != 4){
        printHeader(argv[0]);
        exit(0);
    }

	hostname    = argv[1];
	port        = (atoi(argv[2]) > 0 && atoi(argv[2]) < 65535) ? atoi(argv[2]) : DEFAULT_PORT;
	ctxMethod   = (argc == 4) ? atoi(argv[3]) : 4;  // SSLv2, SSLv3, SSLv2&3 or TLSv1
    ctx         = initSSLContext(ctxMethod);        // load SSL library and dependances
    sock        = makeClientSocket(hostname, port); // make a classic socket to the hostname throught the port
    ssl         = SSL_new(ctx);						// create new SSL connection state



    SSL_set_fd(ssl, sock);				            // attach the socket descriptor

    if(SSL_connect(ssl) == -1)			            // make the SSL connection
        ERR_print_errors_fp(stderr);
    else{
        SSL_set_mode(ssl, SSL_MODE_AUTO_RETRY);     // if the server suddenly wants a new handshake,
                                                    // OpenSSL handles it in the background. Without this
                                                    // option, any read or write operation will return an
                                                    // error if the server wants a new handshake.

        char msg[] = "ClientName";
        printf("[+] Cipher used : %s\n", SSL_get_cipher(ssl));
        showCerts(ssl);								// get any certificats
        SSL_write(ssl, msg, strlen(msg));			// encrypt and send message
        bytes = SSL_read(ssl, buf, sizeof(buf));	// get response and decrypt content
        buf[bytes] = 0;
        printf("[+] Server data received : %s\n", buf);
        SSL_shutdown(ssl);
        SSL_free(ssl);								// release SSL connection state

    }
    CLOSESOCKET(sock);                              // close socket
    #ifdef _WIN32
    WSACleanup();                                   // Windows's Winsock clean
    #endif
    SSL_CTX_free(ctx);								// release SSL's context
    return 0;
}
